# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Running with Scissors in the Matrix (two player, repeated).

Example video: https://youtu.be/rZH9nUKefcU

Players can move around the map and collect resources of `K` discrete types. In
addition to movement, the agents have an action to fire an "interaction" beam.
All players carry an inventory with the count of resources picked up since last
respawn.

Players can observe their own inventory but not the inventories of their
coplayers. When another agent is zapped with the interaction beam, an
interaction occurs. The resolution of the interactions is determined by a
traditional matrix game, where there is a `K x K` payoff matrix describing the
reward produced by the pure strategies available to the two players. The
resources map one-to-one to the pure strategies of the matrix game. Unless
stated otherwise, for the purposes of resolving the interaction, the zapping
agent is considered the row player, and the zapped agent the column player. The
actual strategy played depends on the resources picked up before the
interaction. The more resources of a given type an agent picks up, the more
committed the agent becomes to the pure strategy corresponding to that resource.

In the case of running with scissors, `K = 3`, corresponding to rock, paper, and
scissors pure strategies respectively.

The payoff matrix is the traditional rock-paper-scissors game matrix.

Running with scissors was first described in Vezhnevets et al. (2020). Two
players gather rock, paper or scissor resources in the environment and can
challenge one another to a 'rock, paper scissor' game, the outcome of which
depends on the resources they collected. It is possible to observe the policy
that one's partner is starting to implement, either by watching them pick up
resources or by noting which resources are missing, and then take
countermeasures. This induces a wealth of possible feinting strategies.

Players can also zap resources with their interaction beam to destroy them. This
creates additional scope for feinting strategies.

Players have a `5 x 5` observation window.

The episode has a chance of ending stochastically on every 100 step interval
after step 1000. This usually allows time for 8 or more interactions.

Vezhnevets, A., Wu, Y., Eckstein, M., Leblond, R. and Leibo, J.Z., 2020. OPtions
as REsponses: Grounding behavioural hierarchies in multi-agent reinforcement
learning. In International Conference on Machine Learning (pp. 9733-9742). PMLR.
"""

from typing import Any, Dict, Mapping, Sequence, Tuple

from ml_collections import config_dict

from meltingpot.configs.substrates import the_matrix
from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

# The number of resources must match the (square) size of the matrix.
NUM_RESOURCES = 3

# This color is yellow.
RESOURCE1_COLOR = (255, 227, 11, 255)
RESOURCE1_HIGHLIGHT_COLOR = (255, 214, 91, 255)
RESOURCE1_COLOR_DATA = (RESOURCE1_COLOR, RESOURCE1_HIGHLIGHT_COLOR)
# This color is violet.
RESOURCE2_COLOR = (109, 42, 255, 255)
RESOURCE2_HIGHLIGHT_COLOR = (132, 91, 255, 255)
RESOURCE2_COLOR_DATA = (RESOURCE2_COLOR, RESOURCE2_HIGHLIGHT_COLOR)
# This color is cyan.
RESOURCE3_COLOR = (42, 188, 255, 255)
RESOURCE3_HIGHLIGHT_COLOR = (91, 214, 255, 255)
RESOURCE3_COLOR_DATA = (RESOURCE3_COLOR, RESOURCE3_HIGHLIGHT_COLOR)

ASCII_MAP = """
WWWWWWWWWWWWWWWWWWWWWWW
Wn         n         nW
W   WWW W     W WW    W
W    W rra   app W    W
Wn  WW rra   app WW  nW
W      rra   app      W
W                     W
Wn WW      n         nW
W             WWWW    W
W      ssa W          W
Wn   W ssa W aaa W   nW
W    W ssa W aaa WW   W
W  WWWW W  W  W WWW   W
Wn         n         nW
WWWWWWWWWWWWWWWWWWWWWWW
"""

_resource_names = [
    "resource_class1",
    "resource_class2",
    "resource_class3",
]

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
CHAR_PREFAB_MAP = {
    "a": {"type": "choice", "list": _resource_names},
    "r": _resource_names[0],
    "p": _resource_names[1],
    "s": _resource_names[2],
    "n": "spawn_point",
    "W": "wall",
}

_COMPASS = ["N", "E", "S", "W"]

WALL = {
    "name": "wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall",
                "stateConfigs": [{
                    "state": "wall",
                    "layer": "upperPhysical",
                    "sprite": "Wall",
                }],
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Wall"],
                "spriteShapes": [shapes.WALL],
                "palettes": [{"*": (95, 95, 95, 255),
                              "&": (100, 100, 100, 255),
                              "@": (109, 109, 109, 255),
                              "#": (152, 152, 152, 255)}],
                "noRotates": [False]
            }
        },
        {
            "component": "BeamBlocker",
            "kwargs": {
                "beamType": "gameInteraction"
            }
        },
    ]
}

SPAWN_POINT = {
    "name": "spawnPoint",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "layer": "alternateLogic",
                    "groups": ["spawnPoints"]
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}

# PLAYER_COLOR_PALETTES is a list with each entry specifying the color to use
# for the player at the corresponding index.
NUM_PLAYERS_UPPER_BOUND = 8
PLAYER_COLOR_PALETTES = []
for idx in range(NUM_PLAYERS_UPPER_BOUND):
  PLAYER_COLOR_PALETTES.append(shapes.get_palette(colors.palette[idx]))

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "interact": 0}
FORWARD    = {"move": 1, "turn":  0, "interact": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "interact": 0}
BACKWARD   = {"move": 3, "turn":  0, "interact": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "interact": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "interact": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "interact": 0}
INTERACT   = {"move": 0, "turn":  0, "interact": 1}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    INTERACT,
)

TARGET_SPRITE_SELF = {
    "name": "Self",
    "shape": shapes.CUTE_AVATAR,
    "palette": shapes.get_palette((50, 100, 200)),
    "noRotate": True,
}

TARGET_SPRITE_OTHER = {
    "name": "Other",
    "shape": shapes.CUTE_AVATAR,
    "palette": shapes.get_palette((200, 100, 50)),
    "noRotate": True,
}


def create_scene():
  """Creates the global scene."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "TheMatrix",
              "kwargs": {
                  # Prevent interaction before both interactors have collected
                  # at least one resource.
                  "disallowUnreadyInteractions": True,
                  "matrix": [
                      [0, -10, 10],
                      [10, 0, -10],
                      [-10, 10, 0]
                  ],
                  "resultIndicatorColorIntervals": [
                      (-10.0, -5.0),  # red
                      (-5.0, -2.5),  # yellow
                      (-2.5, 2.5),  # green
                      (2.5, 5.0),  # blue
                      (5.0, 10.0)  # violet
                  ],
              }
          },
          {
              "component": "StochasticIntervalEpisodeEnding",
              "kwargs": {
                  "minimumFramesPerEpisode": 1000,
                  "intervalLength": 100,  # Set equal to unroll length.
                  "probabilityTerminationPerInterval": 0.2
              }
          }
      ]
  }
  return scene


def create_resource_prefab(
    resource_id: int,
    resource_shape: str,
    resource_palette: Dict[str, Tuple[int, int, int, int]]):
  """Creates resource prefab with provided resource_id, shape, and palette."""
  resource_name = "resource_class{}".format(resource_id)
  resource_prefab = {
      "name": resource_name,
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": resource_name,
                  "stateConfigs": [
                      {"state": resource_name + "_wait",
                       "groups": ["resourceWaits"]},
                      {"state": resource_name,
                       "layer": "lowerPhysical",
                       "sprite": resource_name + "_sprite"},
                  ]
              },
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [resource_name + "_sprite"],
                  "spriteShapes": [resource_shape],
                  "palettes": [resource_palette],
                  "noRotates": [True]
              },
          },
          {
              "component": "Resource",
              "kwargs": {
                  "resourceClass": resource_id,
                  "visibleType": resource_name,
                  "waitState": resource_name + "_wait",
                  "regenerationRate": 0.02,
                  "regenerationDelay": 15,
              },
          },
          {
              "component": "Destroyable",
              "kwargs": {
                  "waitState": resource_name + "_wait",
                  # It is possible to destroy resources but takes concerted
                  # effort to do so by zapping them `initialHealth` times.
                  "initialHealth": 3,
              },
          },
      ]
  }
  return resource_prefab


def create_avatar_object(
    player_idx: int,
    all_source_sprite_names: Sequence[str],
    target_sprite_self: Dict[str, Any],
    target_sprite_other: Dict[str, Any],
    turn_off_default_reward: bool = False) -> Dict[str, Any]:
  """Create an avatar object given self vs other sprite data."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  # Setup the self vs other sprite mapping.
  source_sprite_self = "Avatar" + str(lua_index)
  custom_sprite_map = {source_sprite_self: target_sprite_self["name"]}
  for name in all_source_sprite_names:
    if name != source_sprite_self:
      custom_sprite_map[name] = target_sprite_other["name"]

  live_state_name = "player{}".format(lua_index)
  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": live_state_name,
                  "stateConfigs": [
                      {"state": live_state_name,
                       "layer": "upperPhysical",
                       "sprite": source_sprite_self,
                       "contact": "avatar",
                       "groups": ["players"]},

                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "colored_square",
                  "spriteNames": [source_sprite_self],
                  # A white square should never be displayed. It will always be
                  # remapped since this is self vs other observation mode.
                  "spriteRGBColors": [(255, 255, 255, 255)],
              }
          },
          {
              "component": "AdditionalSprites",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "customSpriteNames": [target_sprite_self["name"],
                                        target_sprite_other["name"]],
                  "customSpriteShapes": [target_sprite_self["shape"],
                                         target_sprite_other["shape"]],
                  "customPalettes": [target_sprite_self["palette"],
                                     target_sprite_other["palette"]],
                  "customNoRotates": [target_sprite_self["noRotate"],
                                      target_sprite_other["noRotate"]],
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "aliveState": live_state_name,
                  "waitState": "playerWait",
                  "speed": 1.0,
                  "spawnGroup": "spawnPoints",
                  "actionOrder": ["move", "turn", "interact"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "interact": {"default": 0, "min": 0, "max": 1},
                  },
                  "view": {
                      "left": 2,
                      "right": 2,
                      "forward": 3,
                      "backward": 1,
                      "centered": False
                  },
                  "spriteMap": custom_sprite_map,
                  # The following kwarg makes it possible to get rewarded even
                  # on frames when an avatar is "dead". It is needed for in the
                  # matrix games in order to correctly handle the case of two
                  # players getting hit simultaneously by the same beam.
                  "skipWaitStateRewards": False,
              }
          },
          {
              "component": "GameInteractionZapper",
              "kwargs": {
                  "cooldownTime": 2,
                  "beamLength": 3,
                  "beamRadius": 1,
                  "framesTillRespawn": 5,
                  "numResources": NUM_RESOURCES,
                  "endEpisodeOnFirstInteraction": False,
                  # Reset both players' inventories after each interaction.
                  "reset_winner_inventory": True,
                  "reset_loser_inventory": True,
                  # Both players get removed after each interaction.
                  "losingPlayerDies": True,
                  "winningPlayerDies": True,
                  # `freezeOnInteraction` is the number of frames to display the
                  # interaction result indicator, freeze, and delay delivering
                  # all results of interacting.
                  "freezeOnInteraction": 16,
              }
          },
          {
              "component": "ReadyToShootObservation",
              "kwargs": {
                  "zapperComponent": "GameInteractionZapper",
              }
          },
          {
              "component": "InventoryObserver",
              "kwargs": {
              }
          },
          {
              "component": "SpawnResourcesWhenAllPlayersZapped",
          },
          {
              "component": "Taste",
              "kwargs": {
                  "mostTastyResourceClass": -1,  # -1 indicates no preference.
                  # No resource is most tasty when mostTastyResourceClass == -1.
                  "mostTastyReward": 0.1,
              }
          },
          {
              "component": "InteractionTaste",
              "kwargs": {
                  "mostTastyResourceClass": -1,  # -1 indicates no preference.
                  "zeroDefaultInteractionReward": turn_off_default_reward,
                  "extraReward": 1.0,
              }
          },
          {
              "component": "AvatarMetricReporter",
              "kwargs": {
                  "metrics": [
                      {
                          # Report the inventories of both players involved in
                          # an interaction on this frame formatted as
                          # (self inventory, partner inventory).
                          "name": "INTERACTION_INVENTORIES",
                          "type": "tensor.DoubleTensor",
                          "shape": (2, NUM_RESOURCES),
                          "component": "GameInteractionZapper",
                          "variable": "latest_interaction_inventories",
                      },
                      *the_matrix.get_cumulant_metric_configs(NUM_RESOURCES),
                  ]
              }
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return avatar_object


def create_prefabs():
  """Returns a dictionary mapping names to template game objects."""
  prefabs = {
      "wall": WALL,
      "spawn_point": SPAWN_POINT,
  }
  prefabs["resource_class1"] = create_resource_prefab(
      1, shapes.BUTTON, {"*": RESOURCE1_COLOR_DATA[0],
                         "#": RESOURCE1_COLOR_DATA[1],
                         "x": (0, 0, 0, 0)})
  prefabs["resource_class2"] = create_resource_prefab(
      2, shapes.BUTTON, {"*": RESOURCE2_COLOR_DATA[0],
                         "#": RESOURCE2_COLOR_DATA[1],
                         "x": (0, 0, 0, 0)})
  prefabs["resource_class3"] = create_resource_prefab(
      3, shapes.BUTTON, {"*": RESOURCE3_COLOR_DATA[0],
                         "#": RESOURCE3_COLOR_DATA[1],
                         "x": (0, 0, 0, 0)})
  return prefabs


def get_all_source_sprite_names(num_players):
  all_source_sprite_names = []
  for player_idx in range(0, num_players):
    # Lua is 1-indexed.
    lua_index = player_idx + 1
    all_source_sprite_names.append("Avatar" + str(lua_index))

  return all_source_sprite_names


def create_avatar_objects(num_players,
                          turn_off_default_reward: bool = False):
  """Returns list of avatar objects of length 'num_players'."""
  all_source_sprite_names = get_all_source_sprite_names(num_players)
  avatar_objects = []
  for player_idx in range(0, num_players):
    game_object = create_avatar_object(
        player_idx,
        all_source_sprite_names,
        TARGET_SPRITE_SELF,
        TARGET_SPRITE_OTHER,
        turn_off_default_reward=turn_off_default_reward)
    readiness_marker = the_matrix.create_ready_to_interact_marker(player_idx)
    avatar_objects.append(game_object)
    avatar_objects.append(readiness_marker)

  return avatar_objects


def create_world_sprite_map(
    num_players: int, target_sprite_other: Dict[str, Any]) -> Dict[str, str]:
  all_source_sprite_names = get_all_source_sprite_names(num_players)
  world_sprite_map = {}
  for name in all_source_sprite_names:
    world_sprite_map[name] = target_sprite_other["name"]

  return world_sprite_map


def get_config():
  """Default configuration."""
  config = config_dict.ConfigDict()

  # Other parameters that are useful to override in training config files.
  config.turn_off_default_reward = False

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "INVENTORY",
      "READY_TO_SHOOT",
      # Debug only (do not use the following observations in policies).
      "INTERACTION_INVENTORIES",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(ACTION_SET))
  config.timestep_spec = specs.timestep({
      "RGB": specs.rgb(40, 40),
      "INVENTORY": specs.inventory(3),
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      # Debug only (do not use the following observations in policies).
      "INTERACTION_INVENTORIES": specs.interaction_inventories(3),
      "WORLD.RGB": specs.rgb(120, 184),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"default"})
  config.default_player_roles = ("default",) * 2

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build substrate definition given roles."""
  del config
  num_players = len(roles)
  # Build the rest of the substrate definition.
  substrate_definition = dict(
      levelName="the_matrix",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      # Define upper bound of episode length since episodes end stochastically.
      maxEpisodeLengthFrames=5000,
      spriteSize=8,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": ASCII_MAP,
          "gameObjects": create_avatar_objects(num_players=num_players),
          "scene": create_scene(),
          "prefabs": create_prefabs(),
          "charPrefabMap": CHAR_PREFAB_MAP,
          # worldSpriteMap is needed to make the global view used in videos be
          # be informative in cases where individual avatar views have had
          # sprites remapped to one another (example: self vs other mode).
          "worldSpriteMap": create_world_sprite_map(num_players,
                                                    TARGET_SPRITE_OTHER),
      }
  )
  return substrate_definition



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Puppet policy implementation."""

from typing import Generic, Tuple, TypeVar

import dm_env

from meltingpot.utils.policies import policy
from meltingpot.utils.puppeteers import puppeteer as puppeteer_lib

PuppeteerState = TypeVar('PuppeteerState')
PolicyState = TypeVar('PolicyState')


class PuppetPolicy(policy.Policy[Tuple[PuppeteerState, PolicyState]],
                   Generic[PuppeteerState, PolicyState]):
  """A puppet policy controlled by a puppeteer function."""

  def __init__(
      self,
      puppeteer: puppeteer_lib.Puppeteer[PuppeteerState],
      puppet: policy.Policy[PolicyState]) -> None:
    """Creates a new PuppetBot.

    Args:
      puppeteer: Puppeteer that will be called at every step to modify the
        timestep forwarded to the underlying puppet.
      puppet: The puppet policy. Will be closed with this wrapper.
    """
    self._puppeteer = puppeteer
    self._puppet = puppet

  def step(
      self,
      timestep: dm_env.TimeStep,
      prev_state: Tuple[PuppeteerState, PolicyState],
  ) -> Tuple[int, Tuple[PuppeteerState, PolicyState]]:
    """See base class."""
    puppeteer_state, puppet_state = prev_state
    puppet_timestep, puppeteer_state = self._puppeteer.step(
        timestep, puppeteer_state)
    action, puppet_state = self._puppet.step(puppet_timestep, puppet_state)
    next_state = (puppeteer_state, puppet_state)
    return action, next_state

  def initial_state(self) -> Tuple[PuppeteerState, PolicyState]:
    """See base class."""
    return (self._puppeteer.initial_state(), self._puppet.initial_state())

  def close(self) -> None:
    """See base class."""
    self._puppet.close()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configs for substrates."""

from collections.abc import Mapping, Sequence, Set
import dataclasses
import functools
import importlib
from typing import Any

from ml_collections import config_dict


def _validated(build):
  """And adds validation checks to build function."""

  def lab2d_settings_builder(
      *,
      config: config_dict.ConfigDict,
      roles: Sequence[str],
  ) -> Mapping[str, Any]:
    """Builds the lab2d settings for the specified config and roles.

    Args:
      config: the meltingpot substrate config.
      roles: the role for each corresponding player.

    Returns:
      The lab2d settings for the substrate.
    """
    invalid_roles = set(roles) - config.valid_roles
    if invalid_roles:
      raise ValueError(f'Invalid roles: {invalid_roles!r}. Must be one of '
                       f'{config.valid_roles!r}')
    return build(config=config, roles=roles)

  return lab2d_settings_builder


def get_config(substrate: str) -> config_dict.ConfigDict:
  """Returns the specified config.

  Args:
    substrate: the name of the substrate. Must be in SUBSTRATES.

  Raises:
    ModuleNotFoundError: the config does not exist.
  """
  if substrate not in SUBSTRATES:
    raise ValueError(f'{substrate} not in {SUBSTRATES}.')
  path = f'{__name__}.{substrate}'
  module = importlib.import_module(path)
  config = module.get_config()
  with config.unlocked():
    config.lab2d_settings_builder = _validated(module.build)
  return config.lock()


SUBSTRATES: Set[str] = frozenset({
    # keep-sorted start
    'allelopathic_harvest__open',
    'bach_or_stravinsky_in_the_matrix__arena',
    'bach_or_stravinsky_in_the_matrix__repeated',
    'boat_race__eight_races',
    'chemistry__three_metabolic_cycles',
    'chemistry__three_metabolic_cycles_with_plentiful_distractors',
    'chemistry__two_metabolic_cycles',
    'chemistry__two_metabolic_cycles_with_distractors',
    'chicken_in_the_matrix__arena',
    'chicken_in_the_matrix__repeated',
    'clean_up',
    'coins',
    'collaborative_cooking__asymmetric',
    'collaborative_cooking__circuit',
    'collaborative_cooking__cramped',
    'collaborative_cooking__crowded',
    'collaborative_cooking__figure_eight',
    'collaborative_cooking__forced',
    'collaborative_cooking__ring',
    'commons_harvest__closed',
    'commons_harvest__open',
    'commons_harvest__partnership',
    'coop_mining',
    'daycare',
    'externality_mushrooms__dense',
    'factory_commons__either_or',
    'fruit_market__concentric_rivers',
    'gift_refinements',
    'hidden_agenda',
    'paintball__capture_the_flag',
    'paintball__king_of_the_hill',
    'predator_prey__alley_hunt',
    'predator_prey__open',
    'predator_prey__orchard',
    'predator_prey__random_forest',
    'prisoners_dilemma_in_the_matrix__arena',
    'prisoners_dilemma_in_the_matrix__repeated',
    'pure_coordination_in_the_matrix__arena',
    'pure_coordination_in_the_matrix__repeated',
    'rationalizable_coordination_in_the_matrix__arena',
    'rationalizable_coordination_in_the_matrix__repeated',
    'running_with_scissors_in_the_matrix__arena',
    'running_with_scissors_in_the_matrix__one_shot',
    'running_with_scissors_in_the_matrix__repeated',
    'stag_hunt_in_the_matrix__arena',
    'stag_hunt_in_the_matrix__repeated',
    'territory__inside_out',
    'territory__open',
    'territory__rooms',
    # keep-sorted end
})



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Multi-player environment builder for Melting Pot levels."""

from collections.abc import Mapping
import copy
import itertools
import os
import random
from typing import Any, Optional, Union

from absl import logging
import dmlab2d
from dmlab2d import runfiles_helper
from dmlab2d import settings_helper
from ml_collections import config_dict
import tree

from meltingpot.utils.substrates import game_object_utils
from meltingpot.utils.substrates.wrappers import reset_wrapper

Settings = Union[config_dict.ConfigDict, Mapping[str, Any]]

_MAX_SEED = 2 ** 32 - 1
_DMLAB2D_ROOT = runfiles_helper.find()


def _find_root() -> str:
  import re  # pylint: disable=g-import-not-at-top
  return re.sub('^(.*)/meltingpot/.*?$', r'\1', __file__)


_MELTINGPOT_ROOT = _find_root()


# Although to_dict in ConfigDict is recursive, it is not enough for our use case
# because the recursion will _not_ go into the list elements. And we have plenty
# of those in our configs.
def _config_dict_to_dict(value):
  if isinstance(value, config_dict.ConfigDict):
    return tree.map_structure(_config_dict_to_dict, value.to_dict())
  return value


def parse_python_settings_for_dmlab2d(
    lab2d_settings: config_dict.ConfigDict) -> dict[str, Any]:
  """Flatten lab2d_settings into Lua-friendly properties."""
  # Since config_dicts disallow "." in keys, we must use a different character,
  # "$", in our config and then convert it to "." here. This is particularly
  # important for levels with config keys like 'player.%default' in DMLab2D.
  lab2d_settings = _config_dict_to_dict(lab2d_settings)
  lab2d_settings = settings_helper.flatten_args(lab2d_settings)
  lab2d_settings_dict = {}
  for key, value in lab2d_settings.items():
    converted_key = key.replace("$", ".")
    lab2d_settings_dict[converted_key] = str(value)
  return lab2d_settings_dict


def apply_prefab_overrides(
    lab2d_settings: config_dict.ConfigDict,
    prefab_overrides: Optional[Settings] = None) -> None:
  """Apply prefab overrides to lab2d_settings."""
  if "gameObjects" not in lab2d_settings.simulation:
    lab2d_settings.simulation.gameObjects = []

  # Edit prefabs with the overrides, both in lab2d_settings and in prefabs.
  if prefab_overrides:
    for prefab, override in prefab_overrides.items():
      for component, arg_overrides in override.items():
        for arg_name, arg_override in arg_overrides.items():
          if prefab not in lab2d_settings.simulation.prefabs:
            raise ValueError(f"Prefab override for '{prefab}' given, but not " +
                             "available in `prefabs`.")
          game_object_utils.get_first_named_component(
              lab2d_settings.simulation.prefabs[prefab],
              component)["kwargs"][arg_name] = arg_override


def maybe_build_and_add_avatar_objects(
    lab2d_settings: config_dict.ConfigDict) -> None:
  """If requested, build the avatar objects and add them to lab2d_settings.

  Avatars will be built here if and only if:
  1) An 'avatar' prefab is supplied in lab2d_settings.simulation.prefabs; and
  2) lab2d_settings.simulation.buildAvatars is not True.

  Avatars built here will have their colors set from the palette provided in
  lab2d_settings.simulation.playerPalettes, or if none is provided, using the
  first num_players colors in the colors.py module.

  Args:
    lab2d_settings: A writable version of the lab2d_settings. Avatar objects,
      if they are to be built here, will be added as game objects in
      lab2d_settings.simulation.gameObjects.
  """
  # Whether the avatars will be built in Lua (False) or here (True). This is
  # roughly the opposite of the `buildAvatars` setting.
  build_avatars_here = ("avatar" in lab2d_settings.simulation.prefabs)
  if ("buildAvatars" in lab2d_settings.simulation
      and lab2d_settings.simulation.buildAvatars):
    build_avatars_here = False
    if "avatar" not in lab2d_settings.simulation.prefabs:
      raise ValueError(
          "Deferring avatar building to Lua, yet no 'avatar' prefab given.")
  if build_avatars_here:
    palettes = (lab2d_settings.simulation.playerPalettes
                if "playerPalettes" in lab2d_settings.simulation else None)
    if "gameObjects" not in lab2d_settings.simulation:
      lab2d_settings.simulation.gameObjects = []
    # Create avatars.
    logging.info("Building avatars in `meltingpot.builder` with palettes: %s",
                 lab2d_settings.simulation.playerPalettes)
    avatar_objects = game_object_utils.build_avatar_objects(
        int(lab2d_settings.numPlayers),
        lab2d_settings.simulation.prefabs,
        palettes)
    lab2d_settings.simulation.gameObjects += avatar_objects


def locate_and_overwrite_level_directory(
    lab2d_settings: config_dict.ConfigDict) -> None:
  """Locates the run files, and overwrites the levelDirectory with it."""
  # Locate runfiles.
  level_name = lab2d_settings.get("levelName")
  level_dir = lab2d_settings.get("levelDirectory")
  if level_dir:
    lab2d_settings.levelName = os.path.join(level_dir, level_name)
    lab2d_settings.levelDirectory = _MELTINGPOT_ROOT


def builder(
    lab2d_settings: Settings,
    prefab_overrides: Optional[Settings] = None,
    env_seed: Optional[int] = None,
    **settings) -> dmlab2d.Environment:
  """Builds a Melting Pot environment.

  Args:
    lab2d_settings: a dict of environment designation args.
    prefab_overrides: overrides for prefabs.
    env_seed: the seed to pass to the environment.
    **settings: Other settings which are not used by Melting Pot but can still
      be passed from the environment builder.

  Returns:
    A multi-player Melting Pot environment.
  """
  del settings  #  Not currently used by DMLab2D.

  assert "simulation" in lab2d_settings

  # Copy config, so as not to modify it.
  lab2d_settings = config_dict.ConfigDict(
      copy.deepcopy(lab2d_settings)).unlock()

  apply_prefab_overrides(lab2d_settings, prefab_overrides)
  maybe_build_and_add_avatar_objects(lab2d_settings)
  locate_and_overwrite_level_directory(lab2d_settings)

  # Convert settings from python to Lua format.
  lab2d_settings_dict = parse_python_settings_for_dmlab2d(lab2d_settings)

  if env_seed is None:
    # Select a long seed different than zero.
    env_seed = random.randint(1, _MAX_SEED)
  env_seeds = (seed % (_MAX_SEED + 1) for seed in itertools.count(env_seed))

  def build_environment():
    seed = next(env_seeds)
    lab2d_settings_dict["env_seed"] = str(seed)  # Sets the Lua seed.
    env_raw = dmlab2d.Lab2d(_DMLAB2D_ROOT, lab2d_settings_dict)
    observation_names = env_raw.observation_names()
    return dmlab2d.Environment(
        env=env_raw,
        observation_names=observation_names,
        seed=seed)

  # Add a wrapper that rebuilds the environment when reset is called.
  env = reset_wrapper.ResetWrapper(build_environment)

  return env



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common configuration for all predator_prey__* substrates.

There are two roles: predators and prey. The prey try to eat apples and acorns.
The predators try to eat prey.

Apples are worth 1 reward and can be eaten immediately. Acorns are worth 18
reward but they take a long time to eat. It is not possible to move while eating
so a prey player is especially vulnerable while they eat it.

Predators can also eat other predators, though they get no reward for doing so.
However, a predator might eat another predator anyway in order to remove a
competitor who might otherwise eat its prey.

When prey travel together in tight groups they can defend themselves from being
eaten by predators. When a predator tries to eat its prey then all other prey
who are not currently eating an acorn within a radius of 3 are counted. If there
are more prey than predators within the radius then the predator cannot eat the
prey.

So prey are safer in groups. However, they are also tempted to depart from their
group and strike out alone since that way they are more likely to be the one to
come first to any food they find.

Both predators and prey have limited stamina. They can only move at top speed
for a limited number of consecutive steps, after which they must slow down.
Stamina is visible to all with a colored bar above each player's head. If the
bar over a particular player's head is invisible or green then they can move at
top speed. If it is red then they have depleted their stamina and can only move
slowly until they let it recharge. Stamina is recharged by standing still for
some number of consecutive time steps, how many depends on how much stamina was
depleted. Predators have a faster top speed than prey but they tire more
quickly.

Prey but not predators can cross tall grass (green grassy locations). Prey must
still be careful on grass though since predators can still reach one cell over
the border to eat prey on the edge of safety.

Both predators and prey respawn 200 steps after being eaten.
"""

from typing import Any, Dict, Generator, Mapping, Sequence

from ml_collections import config_dict as configdict

from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

_COMPASS = ("N", "E", "S", "W")
ITEMS = ("empty", "acorn")
INVISIBLE = (0, 0, 0, 0)

SPRITES = {}
PALETTES = {}

SPRITES["empty"] = """
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
"""

PALETTES["empty"] = {"x": INVISIBLE,}

SPRITES["acorn"] = """
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxBB
xxxxxxbb
xxxxxxxx
"""

PALETTES["acorn"] = {
    "x": INVISIBLE,
    "B": [158, 85, 25, 255],
    "b": [92, 29, 19, 255],
}

PREDATOR_EAT_SPRITE = """
*x*x*x*x
*x*x*x**
x***&x**
xx*&xx**
xx*&xx*&
xx*&xx*x
xx&&xx&x
xxxxxxxx
"""

APPLE_SPRITE = """
xxxxxxxx
xxo|*xxx
x*#|**xx
x#***#xx
x#####xx
xx###xxx
xxxxxxxx
xxxxxxxx
"""


def create_inventory(player_index: int):
  """Return prefab for the inventory of the player at index `player_index`."""
  lua_idx = player_index + 1
  prefab = {
      "name": "inventory",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "wait",
                  "stateConfigs": (
                      [{"state": "wait"}] +
                      [{"state": item, "sprite": item, "layer": "overlay"}
                       for item in ITEMS]),
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ITEMS,
                  "spriteShapes": [SPRITES[item] for item in ITEMS],
                  "palettes": [PALETTES[item] for item in ITEMS],
                  "noRotates": [False]
              }
          },
          {
              "component": "AvatarConnector",
              "kwargs": {
                  "playerIndex": lua_idx,
                  "aliveState": "empty",
                  "waitState": "wait"
              }
          },
          {
              "component": "Inventory",
              "kwargs": {
                  "playerIndex": lua_idx,
              }
          },
      ]
  }
  return prefab


def create_base_prefab(name, layer="upperPhysical"):
  """Returns a base prefab with a given name on the given layer."""
  return {
      "name": f"{name}",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": f"{name}",
                  "stateConfigs": [{
                      "state": f"{name}",
                      "layer": layer,
                      "sprite": f"{name}",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [name],
                  "spriteShapes": [SPRITES[name]],
                  "palettes": [PALETTES[name]],
                  "noRotates": [True]
              }
          }]
  }

NW_WALL_CORNER = {
    "name": "nw_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "nw_wall_corner",
                "stateConfigs": [{
                    "state": "nw_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "NwWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NwWallCorner"],
                "spriteShapes": [shapes.BRICK_WALL_NW_CORNER],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

NE_WALL_CORNER = {
    "name": "ne_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "ne_wall_corner",
                "stateConfigs": [{
                    "state": "ne_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "NeWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NeWallCorner"],
                "spriteShapes": [shapes.BRICK_WALL_NE_CORNER],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SE_WALL_CORNER = {
    "name": "se_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "se_wall_corner",
                "stateConfigs": [{
                    "state": "se_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "SeWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SeWallCorner"],
                "spriteShapes": [shapes.BRICK_WALL_SE_CORNER],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SW_WALL_CORNER = {
    "name": "sw_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "sw_wall_corner",
                "stateConfigs": [{
                    "state": "sw_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "SwWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SwWallCorner"],
                "spriteShapes": [shapes.BRICK_WALL_SW_CORNER],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

NW_INNER_WALL_CORNER = {
    "name": "nw_inner_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "nw_inner_wall_corner",
                "stateConfigs": [{
                    "state": "nw_inner_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "NwInnerWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NwInnerWallCorner"],
                "spriteShapes": [shapes.BRICK_WALL_INNER_NW_CORNER],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

NE_INNER_WALL_CORNER = {
    "name": "ne_inner_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "ne_inner_wall_corner",
                "stateConfigs": [{
                    "state": "ne_inner_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "NeInnerWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NeInnerWallCorner"],
                "spriteShapes": [shapes.BRICK_WALL_INNER_NE_CORNER],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SE_INNER_WALL_CORNER = {
    "name": "se_inner_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "se_inner_wall_corner",
                "stateConfigs": [{
                    "state": "se_inner_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "SeInnerWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SeInnerWallCorner"],
                "spriteShapes": [shapes.BRICK_WALL_INNER_SE_CORNER],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SW_INNER_WALL_CORNER = {
    "name": "sw_inner_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "sw_inner_wall_corner",
                "stateConfigs": [{
                    "state": "sw_inner_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "SwInnerWallCorner",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SwInnerWallCorner"],
                "spriteShapes": [shapes.BRICK_WALL_INNER_SW_CORNER],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

WALL_NORTH = {
    "name": "wall_north",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall_north",
                "stateConfigs": [{
                    "state": "wall_north",
                    "layer": "upperPhysical",
                    "sprite": "WallNorth",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallNorth"],
                "spriteShapes": [shapes.BRICK_WALL_NORTH],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

WALL_EAST = {
    "name": "wall_east",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall_east",
                "stateConfigs": [{
                    "state": "wall_east",
                    "layer": "upperPhysical",
                    "sprite": "WallEast",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallEast"],
                "spriteShapes": [shapes.BRICK_WALL_EAST],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

WALL_SOUTH = {
    "name": "wall_south",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall_south",
                "stateConfigs": [{
                    "state": "wall_south",
                    "layer": "upperPhysical",
                    "sprite": "WallSouth",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallSouth"],
                "spriteShapes": [shapes.BRICK_WALL_SOUTH],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}


WALL_WEST = {
    "name": "wall_west",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall_west",
                "stateConfigs": [{
                    "state": "wall_west",
                    "layer": "upperPhysical",
                    "sprite": "WallWest",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallWest"],
                "spriteShapes": [shapes.BRICK_WALL_WEST],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

FILL = {
    "name": "fill",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "fill",
                "stateConfigs": [{
                    "state": "fill",
                    "layer": "upperPhysical",
                    "sprite": "fill",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["fill"],
                "spriteShapes": [shapes.FILL],
                "palettes": [shapes.BRICK_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

TILED_FLOOR = {
    "name": "tiled_floor",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "tiled_floor",
                "stateConfigs": [{
                    "state": "tiled_floor",
                    "layer": "background",
                    "sprite": "tiled_floor",
                }],
            }
        },
        {"component": "Transform"},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["tiled_floor"],
                "spriteShapes": [shapes.TILED_FLOOR_GREY],
                "palettes": [{"o": (204, 199, 192, 255),
                              "-": (194, 189, 182, 255),}],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS = {
    "name":
        "safe_grass",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass",
                "stateConfigs": [{
                    "state": "safe_grass",
                    "layer": "midPhysical",
                    "sprite": "safe_grass",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass"],
                "spriteShapes": [shapes.GRASS_STRAIGHT],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS_N_EDGE = {
    "name":
        "safe_grass_n_edge",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass_n_edge",
                "stateConfigs": [{
                    "state": "safe_grass_n_edge",
                    "layer": "midPhysical",
                    "sprite": "safe_grass_n_edge",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass_n_edge"],
                "spriteShapes": [shapes.GRASS_STRAIGHT_N_EDGE],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS_E_EDGE = {
    "name":
        "safe_grass_e_edge",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass_e_edge",
                "stateConfigs": [{
                    "state": "safe_grass_e_edge",
                    "layer": "midPhysical",
                    "sprite": "safe_grass_e_edge",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass_e_edge"],
                "spriteShapes": [shapes.GRASS_STRAIGHT_E_EDGE],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS_S_EDGE = {
    "name":
        "safe_grass_s_edge",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass_s_edge",
                "stateConfigs": [{
                    "state": "safe_grass_s_edge",
                    "layer": "midPhysical",
                    "sprite": "safe_grass_s_edge",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass_s_edge"],
                "spriteShapes": [shapes.GRASS_STRAIGHT_S_EDGE],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS_W_EDGE = {
    "name":
        "safe_grass_w_edge",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass_w_edge",
                "stateConfigs": [{
                    "state": "safe_grass_w_edge",
                    "layer": "midPhysical",
                    "sprite": "safe_grass_w_edge",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass_w_edge"],
                "spriteShapes": [shapes.GRASS_STRAIGHT_W_EDGE],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS_NW_CORNER = {
    "name":
        "safe_grass_nw_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass_nw_corner",
                "stateConfigs": [{
                    "state": "safe_grass_nw_corner",
                    "layer": "midPhysical",
                    "sprite": "safe_grass_nw_corner",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass_nw_corner"],
                "spriteShapes": [shapes.GRASS_STRAIGHT_NW_CORNER],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS_NE_CORNER = {
    "name":
        "safe_grass_ne_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass_ne_corner",
                "stateConfigs": [{
                    "state": "safe_grass_ne_corner",
                    "layer": "midPhysical",
                    "sprite": "safe_grass_ne_corner",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass_ne_corner"],
                "spriteShapes": [shapes.GRASS_STRAIGHT_NE_CORNER],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS_SE_CORNER = {
    "name":
        "safe_grass_se_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass_se_corner",
                "stateConfigs": [{
                    "state": "safe_grass_se_corner",
                    "layer": "midPhysical",
                    "sprite": "safe_grass_se_corner",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass_se_corner"],
                "spriteShapes": [shapes.GRASS_STRAIGHT_SE_CORNER],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

SAFE_GRASS_SW_CORNER = {
    "name":
        "safe_grass_sw_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "safe_grass_sw_corner",
                "stateConfigs": [{
                    "state": "safe_grass_sw_corner",
                    "layer": "midPhysical",
                    "sprite": "safe_grass_sw_corner",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["safe_grass_sw_corner"],
                "spriteShapes": [shapes.GRASS_STRAIGHT_SW_CORNER],
                "palettes": [shapes.GRASS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}


def create_apple(apple_reward: float = 1.0):
  """Return a prefab object defining an apple, which can be eaten by prey."""
  prefab = {
      "name":
          "edibleApple",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState":
                      "apple",
                  "stateConfigs": [
                      {
                          "state": "apple",
                          "layer": "lowerPhysical",
                          "sprite": "apple",
                      },
                      {
                          "state": "appleWait",
                          "layer": "logic",
                      },
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["apple"],
                  "spriteShapes": [APPLE_SPRITE],
                  "palettes": [{
                      "*": (106, 184, 83, 255),
                      "#": (96, 166, 75, 255),
                      "o": (61, 130, 62, 255),
                      "|": (115, 62, 57, 255),
                      "x": INVISIBLE,
                  }],
                  "noRotates": [True],
              }
          },
          {
              "component": "AppleEdible",
              "kwargs": {
                  "liveState": "apple",
                  "waitState": "appleWait",
                  "rewardForEating": apple_reward,
              }
          },
          {
              "component": "FixedRateRegrow",
              "kwargs": {
                  "name": "AppleFixedRateRegrow",
                  "liveState": "apple",
                  "waitState": "appleWait",
                  "regrowRate": 0.007,
              }
          },
      ]
  }
  return prefab

FLOOR_ACORN = {
    "name":
        "floorAcorn",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "floorAcorn",
                "stateConfigs": [
                    {
                        "state": "floorAcorn",
                        "layer": "lowerPhysical",
                        "sprite": "floorAcorn",
                    },
                    {
                        "state": "acornWait",
                        "layer": "logic",
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["floorAcorn"],
                "spriteShapes": [shapes.ACORN],
                "palettes": [{
                    "*": (158, 85, 25, 255),
                    "@": (158, 85, 25, 140),
                    "o": (92, 29, 19, 255),
                    "x": INVISIBLE,
                }],
                "noRotates": [True],
            }
        },
        {
            "component": "AcornPickUppable",
            "kwargs": {
                "liveState": "floorAcorn",
                "waitState": "acornWait",
            }
        },
        {
            "component": "FixedRateRegrow",
            "kwargs": {
                "name": "AcornFixedRateRegrow",
                "liveState": "floorAcorn",
                "waitState": "acornWait",
                "regrowRate": 0.01,
            }
        },
    ]
}

SPAWN_POINT = {
    "name": "spawnPoint",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "layer": "alternateLogic",
                    "groups": ["spawnPoints"]
                }],
            }
        },
        {"component": "Transform"},
    ]
}


def create_spawn_point_prefab(team):
  """Return a team-specific spawn-point prefab."""
  prefab = {
      "name": "spawn_point",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "playerSpawnPoint",
                  "stateConfigs": [{
                      "state": "playerSpawnPoint",
                      "layer": "alternateLogic",
                      "groups": ["{}SpawnPoints".format(team)],
                  }],
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "invisible",
                  "spriteNames": [],
                  "spriteRGBColors": []
              }
          },
      ]
  }
  return prefab


def create_prefabs(apple_reward: float = 1.0):
  """Returns the prefabs dictionary."""
  prefabs = {
      "spawn_point_prey": create_spawn_point_prefab("prey"),
      "spawn_point_predator": create_spawn_point_prefab("predator"),
      "apple": create_apple(apple_reward=apple_reward),
      "floor_acorn": FLOOR_ACORN,
      "nw_wall_corner": NW_WALL_CORNER,
      "ne_wall_corner": NE_WALL_CORNER,
      "se_wall_corner": SE_WALL_CORNER,
      "sw_wall_corner": SW_WALL_CORNER,
      "nw_inner_wall_corner": NW_INNER_WALL_CORNER,
      "ne_inner_wall_corner": NE_INNER_WALL_CORNER,
      "se_inner_wall_corner": SE_INNER_WALL_CORNER,
      "sw_inner_wall_corner": SW_INNER_WALL_CORNER,
      "wall_north": WALL_NORTH,
      "wall_east": WALL_EAST,
      "wall_south": WALL_SOUTH,
      "wall_west": WALL_WEST,
      "fill": FILL,
      "tiled_floor": TILED_FLOOR,
      "safe_grass": SAFE_GRASS,
      "safe_grass_n_edge": SAFE_GRASS_N_EDGE,
      "safe_grass_e_edge": SAFE_GRASS_E_EDGE,
      "safe_grass_s_edge": SAFE_GRASS_S_EDGE,
      "safe_grass_w_edge": SAFE_GRASS_W_EDGE,
      "safe_grass_ne_corner": SAFE_GRASS_NE_CORNER,
      "safe_grass_se_corner": SAFE_GRASS_SE_CORNER,
      "safe_grass_sw_corner": SAFE_GRASS_SW_CORNER,
      "safe_grass_nw_corner": SAFE_GRASS_NW_CORNER,
  }
  return prefabs

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "interact": 0}
FORWARD    = {"move": 1, "turn":  0, "interact": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "interact": 0}
BACKWARD   = {"move": 3, "turn":  0, "interact": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "interact": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "interact": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "interact": 0}
INTERACT   = {"move": 0, "turn":  0, "interact": 1}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    INTERACT,
)


def create_scene():
  """Creates the global scene."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {"component": "Transform"},
      ]
  }

  return scene


def _create_avatar_object(player_idx: int, is_predator: bool,
                          max_stamina: int) -> Dict[str, Any]:
  """Create an avatar object."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1
  live_state_name = "player{}".format(lua_index)
  avatar_sprite_name = "avatarSprite{}".format(lua_index)

  if is_predator:
    spawn_group = "predatorSpawnPoints"
    color_palette = shapes.PRED1_PALETTE
    sprite = shapes.PERSISTENCE_PREDATOR
  if not is_predator:
    spawn_group = "preySpawnPoints"
    alert_state_name = live_state_name + "alert"
    sit_state_name = live_state_name + "sit"
    prep_to_eat_state_name = live_state_name + "prepToEat"
    first_bite_state_name = live_state_name + "firstBite"
    second_bite_state_name = live_state_name + "secondBite"
    last_bite_state_name = live_state_name + "lastBite"
    alert_sprite_name = "avatarAlertSprite{}".format(lua_index)
    sit_sprite_name = "avatarSitSprite{}".format(lua_index)
    prep_to_eat_sprite_name = "avatarPrepToEatSprite{}".format(lua_index)
    first_bite_sprite_name = "avatarFirstBiteSprite{}".format(lua_index)
    second_bite_sprite_name = "avatarSecondBiteSprite{}".format(lua_index)
    last_bite_sprite_name = "avatarLastBiteSprite{}".format(lua_index)
    color_palette = {**shapes.get_palette(colors.palette[player_idx]
                                          ), ** PALETTES["acorn"]}
    sprite = shapes.CUTE_AVATAR
    alert_sprite = shapes.CUTE_AVATAR_ALERT
    sit_sprite = shapes.CUTE_AVATAR_SIT
    prep_to_eat_sprite = shapes.CUTE_AVATAR_EAT
    first_bite_sprite = shapes.CUTE_AVATAR_FIRST_BITE
    second_bite_sprite = shapes.CUTE_AVATAR_SECOND_BITE
    last_bite_sprite = shapes.CUTE_AVATAR_LAST_BITE

  interact_palette = {
      "P": colors.palette[player_idx] + (255,),
      "&": (10, 10, 10, 50),
      "*": (230, 230, 230, 255),
      "x": INVISIBLE,
  }

  if is_predator:
    role_name = "predator"
    green_freeze_time = 0
    yellow_freeze_time = 1
    red_freeze_time = 6
  else:
    role_name = "prey"
    green_freeze_time = 1
    yellow_freeze_time = 2
    red_freeze_time = 4

  live_state_name = "player{}".format(lua_index)
  avatar_object = {
      "name": f"avatar{lua_index}",
      "components": [
          {
              "component": "Transform",
          },
          {
              "component": "Role",
              "kwargs": {
                  "isPredator": is_predator,
              }
          },
          {
              "component": "Stamina",
              "kwargs": {
                  "maxStamina": max_stamina,
                  "classConfig": {"name": role_name,
                                  "greenFreezeTime": green_freeze_time,
                                  "yellowFreezeTime": yellow_freeze_time,
                                  "redFreezeTime": red_freeze_time},
                  "amountInvisible": 6,
                  "amountGreen": 6,
                  "amountYellow": 6,
                  "amountRed": 1,
                  "costlyActions": ["move", "turn", "interact"],
              }
          },
          {
              "component": "StaminaObservation",
              "kwargs": {
                  "staminaComponent": "Stamina",
              }
          },
          # The `RewardForStaminaLevel` component defines a pseudoreward which
          # is useful for training background populations to rapidly learn how
          # to control their stamina. For the default "real" substrate, the
          # reward it defines should always be zero.
          {
              "component": "RewardForStaminaLevel",
              "kwargs": {
                  "rewardValue": 0.0,
                  "bands": [],
              }
          },
      ]
  }
  if is_predator:
    avatar_object["components"].extend([
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": live_state_name,
                "stateConfigs": [
                    # Initial player state.
                    {
                        "state": live_state_name,
                        "layer": "upperPhysical",
                        "sprite": avatar_sprite_name,
                        "contact": "avatar",
                        "groups": ["players"]
                    },
                    # Player wait type for times when they are zapped out.
                    {
                        "state": "playerWait",
                        "groups": ["playerWaits"]
                    },
                ]
            }
        },
        {
            "component": "PredatorInteractBeam",
            "kwargs": {
                "cooldownTime": 5,
                "shapes": [PREDATOR_EAT_SPRITE, shapes.FILL],
                "palettes": [interact_palette] * 2,
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": [avatar_sprite_name],
                "spriteShapes": [sprite],
                "palettes": [color_palette],
                "noRotates": [True]
            }
        },
        {
            "component": "Avatar",
            "kwargs": {
                "index": lua_index,
                "aliveState": live_state_name,
                "additionalLiveStates": [],
                "waitState": "playerWait",
                "spawnGroup": spawn_group,
                "actionOrder": ["move",
                                "turn",
                                "interact"],
                "actionSpec": {
                    "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                    "turn": {"default": 0, "min": -1, "max": 1},
                    "interact": {"default": 0, "min": 0, "max": 1},
                },
                "view": {
                    "left": 5,
                    "right": 5,
                    "forward": 9,
                    "backward": 1,
                    "centered": False
                },
            }
        },
        {
            "component": "AvatarEdible",
        },
        {
            "component": "AvatarRespawn",
            "kwargs": {
                "framesTillRespawn": 200,
            }
        },
        ])
  if not is_predator:
    avatar_object["components"].extend([
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": live_state_name,
                "stateConfigs": [
                    # Initial player state.
                    {
                        "state": live_state_name,
                        "layer": "upperPhysical",
                        "sprite": avatar_sprite_name,
                        "contact": "avatar",
                        "groups": ["players"]
                    },
                    # Player wait type for times when they are zapped out.
                    {
                        "state": "playerWait",
                        "groups": ["playerWaits"]
                    },
                    {
                        "state": alert_state_name,
                        "layer": "upperPhysical",
                        "sprite": alert_sprite_name,
                        "contact": "avatar",
                        "groups": ["players"]
                    },
                    {
                        "state": sit_state_name,
                        "layer": "upperPhysical",
                        "sprite": sit_sprite_name,
                        "contact": "avatar",
                        "groups": ["players"]
                    },
                    {
                        "state": prep_to_eat_state_name,
                        "layer": "upperPhysical",
                        "sprite": prep_to_eat_sprite_name,
                        "contact": "avatar",
                        "groups": ["players"]
                    },
                    {
                        "state": first_bite_state_name,
                        "layer": "upperPhysical",
                        "sprite": first_bite_sprite_name,
                        "contact": "avatar",
                        "groups": ["players"]
                    },
                    {
                        "state": second_bite_state_name,
                        "layer": "upperPhysical",
                        "sprite": second_bite_sprite_name,
                        "contact": "avatar",
                        "groups": ["players"]
                    },
                    {
                        "state": last_bite_state_name,
                        "layer": "upperPhysical",
                        "sprite": last_bite_sprite_name,
                        "contact": "avatar",
                        "groups": ["players"]
                    }
                ]
            }
        },
        {
            "component": "Avatar",
            "kwargs": {
                "index": lua_index,
                "aliveState": live_state_name,
                "additionalLiveStates": [alert_state_name,
                                         sit_state_name,
                                         prep_to_eat_state_name,
                                         first_bite_state_name,
                                         second_bite_state_name,
                                         last_bite_state_name],
                "waitState": "playerWait",
                "spawnGroup": spawn_group,
                "actionOrder": ["move", "turn", "interact"],
                "actionSpec": {
                    "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                    "turn": {"default": 0, "min": -1, "max": 1},
                    "interact": {"default": 0, "min": 0, "max": 1},
                    },
                "view": {
                    "left": 5,
                    "right": 5,
                    "forward": 9,
                    "backward": 1,
                    "centered": False
                    },
                }
        },
        {
            "component": "InteractEatAcorn",
            "kwargs": {
                "cooldownTime": 5,
                "shapes": [PREDATOR_EAT_SPRITE, shapes.FILL],
                "palettes": [interact_palette],
                "isEating": False,
                "defaultState": live_state_name,
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": [avatar_sprite_name, alert_sprite_name,
                                sit_sprite_name, prep_to_eat_sprite_name,
                                first_bite_sprite_name, second_bite_sprite_name,
                                last_bite_sprite_name],
                "spriteShapes": [sprite, alert_sprite, sit_sprite,
                                 prep_to_eat_sprite, first_bite_sprite,
                                 second_bite_sprite, last_bite_sprite],
                "palettes": [color_palette] * 7,
                "noRotates": [True] * 7
            }
        },
        {
            "component": "AvatarEatingAnimation",
            "kwargs": {
                "sit": sit_state_name,
                "prepToEat": prep_to_eat_state_name,
                "firstBite": first_bite_state_name,
                "secondBite": second_bite_state_name,
                "lastBite": last_bite_state_name,
                "downState": live_state_name,
                # On each of 3 eating frames, get one third of `acornReward`.
                "acornReward": 18,
                }
        },
        {
            "component": "AvatarEdible",
            "kwargs": {
                "groupRadius": 3,
                "predatorRewardForEating": 1.0,
            }
        },
        {
            "component": "AvatarRespawn",
            "kwargs": {
                "framesTillRespawn": 200,
            }
        },
        {
            "component": "AvatarAnimation",
            "kwargs": {
                "upState": alert_state_name,
                "downState": live_state_name,
            }
        },
        # The `AcornTaste` component defines pseudorewards which are useful for
        # training background populations. For the default "real" substrate,
        # the rewards it defines should always be zero,
        {
            "component": "AcornTaste",
            "kwargs": {
                "collectReward": 0.0,
                "eatReward": 0.0,
                "safeAcornConsumptionReward": 0.0,
            }
        },
    ])

  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return avatar_object


def _create_predator_obstacle(player_idx: int) -> Dict[str, Any]:
  # Lua is 1-indexed.
  lua_idx = player_idx + 1
  return {
      "name":
          "predator_obstacle",
      "components": [
          {
              "component": "StateManager",
              "kwargs":
                  {
                      "initialState": "obstacleWait",
                      "stateConfigs": [
                          {
                              "state": "obstacleWait"
                          },
                          # Block predators from entering any tile with a
                          # piece on layer 'midPhysical'.
                          {
                              "state": "obstacleLive",
                              "layer": "midPhysical",
                              "groups": ["obstacles"]
                          }
                      ]
                  }
          },
          {
              "component": "Transform",
          },
          {
              "component": "AvatarConnector",
              "kwargs": {
                  "playerIndex": lua_idx,
                  "aliveState": "obstacleLive",
                  "waitState": "obstacleWait"
              }
          },
      ]
  }


def _create_stamina_overlay(player_idx: int,
                            max_stamina_bar_states: int,
                            ) -> Generator[Dict[str, Any], None, None]:
  """Create stamina marker overlay objects."""
  # Lua is 1-indexed.
  lua_idx = player_idx + 1

  stamina_bar_state_configs = [
      # Invisible inactive (dead) overlay type.
      {"state": "staminaBarWait"},
  ]
  stamina_bar_sprite_names = []
  stamina_bar_sprite_shapes = []

  # Each player's stamina bars must be in their own layer so they do not
  # interact/collide with other players' stamina bars.
  stamina_bar_layer = f"superOverlay_{player_idx}"

  # Declare one state per level of the stamina bar.
  for i in range(max_stamina_bar_states):
    sprite_name = f"sprite_for_level_{i}"
    stamina_bar_state_configs.append(
        {"state": f"level_{i}",
         "layer": stamina_bar_layer,
         "sprite": sprite_name})
    stamina_bar_sprite_names.append(sprite_name)
    xs = "\nxxxxxxxx"
    blank_space = xs * 7
    number_of_rs = max(6 - i, 0)
    number_of_ys = i if i < 7 else 12 - i
    number_of_gs = max(i - 6, 0)
    if i >= 13:
      level = blank_space + xs
    else:
      level = (
          blank_space
          + "\nx"
          + "G" * number_of_gs
          + "Y" * number_of_ys
          + "R" * number_of_rs
          + "x"
      )
    empty = "\n".join(["x" * 8] * 8)
    # Replace the east/south/west sprites with invisible sprites so the only
    # stamina bar rendered is the one in the direction that the current player
    # is facing.
    stamina_bar_sprite_shapes.append((level, empty, empty, empty))

  # Create a stamina bar for each compass direction. Only the direction the
  # current player is facing is visible.
  for direction in ("N", "E", "S", "W"):
    yield {
        "name": "avatar_stamina_bar",
        "components": [
            {
                "component": "StateManager",
                "kwargs": {
                    "initialState": "staminaBarWait",
                    "stateConfigs": stamina_bar_state_configs
                }
            },
            {
                "component": "Transform",
            },
            {
                "component": "Appearance",
                "kwargs": {
                    "renderMode": "ascii_shape",
                    "spriteNames": stamina_bar_sprite_names,
                    "spriteShapes": stamina_bar_sprite_shapes,
                    "palettes": [{"G": (62, 137, 72, 255),
                                  "Y": (255, 216, 97, 255),
                                  "R": (162, 38, 51, 255),
                                  "x": INVISIBLE,}] * max_stamina_bar_states,
                    "noRotates": [True] * max_stamina_bar_states
                }
            },
            {
                "component": "StaminaBar",
                "kwargs": {
                    "playerIndex": lua_idx,
                    "waitState": "staminaBarWait",
                    "layer": stamina_bar_layer,
                    "direction": direction
                }
            },
        ]
    }


def _build_prey_objects(player_idx: int,
                        max_stamina_bar_states: int = 19):
  """Build a prey avatar and its associated stamina objects."""
  avatar_object = _create_avatar_object(
      player_idx, is_predator=False, max_stamina=max_stamina_bar_states - 1)
  stamina_bar_objects = _create_stamina_overlay(player_idx,
                                                max_stamina_bar_states)
  inventory_object = create_inventory(player_index=player_idx)
  game_objects = []
  game_objects.append(avatar_object)
  game_objects.extend(stamina_bar_objects)
  game_objects.append(inventory_object)
  return game_objects


def _build_predator_objects(player_idx: int,
                            max_stamina_bar_states: int = 19):
  """Build a predator avatar and its associated stamina objects and obstacle."""
  avatar_object = _create_avatar_object(
      player_idx, is_predator=True, max_stamina=max_stamina_bar_states - 1)
  stamina_bar_objects = _create_stamina_overlay(player_idx,
                                                max_stamina_bar_states)
  predator_obstacle = _create_predator_obstacle(player_idx)
  game_objects = []
  game_objects.append(avatar_object)
  game_objects.extend(stamina_bar_objects)
  game_objects.append(predator_obstacle)
  return game_objects


def get_config():
  """Default configuration."""
  config = configdict.ConfigDict()

  # Declare parameters here that we may want to override externally.
  # `apple_reward` should be 1.0 for the canonical version of this environment,
  # but to train background bots it is sometimes useful to use other values in
  # order to control the relative attractiveness of apples and acorns.
  config.apple_reward = 1.0

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "STAMINA",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(ACTION_SET))

  # The roles assigned to each player.
  config.valid_roles = frozenset({"predator", "prey"})

  return config


def build(
    roles: Sequence[str],
    config: configdict.ConfigDict,
) -> Mapping[str, Any]:
  """Build predator_and_prey substrate given player roles."""
  # Build avatars.
  num_players = len(roles)
  avatar_objects_and_helpers = []
  for player_idx, role in enumerate(roles):
    if role == "prey":
      avatar_objects_and_helpers.extend(_build_prey_objects(player_idx))
    elif role == "predator":
      avatar_objects_and_helpers.extend(_build_predator_objects(player_idx))
    else:
      raise ValueError(f"Unrecognized role: {role}")

  # Build the rest of the substrate definition.
  substrate_definition = dict(
      levelName="predator_prey",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      maxEpisodeLengthFrames=1000,
      spriteSize=8,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation=dict(
          map=config.layout.ascii_map,
          gameObjects=avatar_objects_and_helpers,
          scene=create_scene(),
          prefabs=create_prefabs(apple_reward=config.apple_reward),
          charPrefabMap=config.layout.char_prefab_map,
      ),
  )

  return substrate_definition



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Substrate builder."""

from collections.abc import Collection, Mapping, Sequence
from typing import Any

import chex
import dm_env
import reactivex
from reactivex import subject

from meltingpot.utils.substrates import builder
from meltingpot.utils.substrates.wrappers import base
from meltingpot.utils.substrates.wrappers import collective_reward_wrapper
from meltingpot.utils.substrates.wrappers import discrete_action_wrapper
from meltingpot.utils.substrates.wrappers import multiplayer_wrapper
from meltingpot.utils.substrates.wrappers import observables
from meltingpot.utils.substrates.wrappers import observables_wrapper


@chex.dataclass(frozen=True)
class SubstrateObservables:
  """Observables for a substrate.

  Attributes:
    action: emits actions sent to the substrate from players.
    timestep: emits timesteps sent from the substrate to players.
    events: emits environment-specific events resulting from any interactions
      with the Substrate. Each individual event is emitted as a single element:
      (event_name, event_item).
    dmlab2d: Observables from the underlying dmlab2d environment.
  """
  action: reactivex.Observable[Sequence[int]]
  timestep: reactivex.Observable[dm_env.TimeStep]
  events: reactivex.Observable[tuple[str, Any]]
  dmlab2d: observables.Lab2dObservables


class Substrate(base.Lab2dWrapper):
  """Specific subclass of Wrapper with overridden spec types."""

  def __init__(self, env: observables.ObservableLab2d) -> None:
    """See base class."""
    super().__init__(env)
    self._action_subject = subject.Subject()
    self._timestep_subject = subject.Subject()
    self._events_subject = subject.Subject()
    self._observables = SubstrateObservables(
        action=self._action_subject,
        events=self._events_subject,
        timestep=self._timestep_subject,
        dmlab2d=env.observables(),
    )

  def reset(self) -> dm_env.TimeStep:
    """See base class."""
    timestep = super().reset()
    self._timestep_subject.on_next(timestep)
    for event in super().events():
      self._events_subject.on_next(event)
    return timestep

  def step(self, action: Sequence[int]) -> dm_env.TimeStep:
    """See base class."""
    self._action_subject.on_next(action)
    timestep = super().step(action)
    self._timestep_subject.on_next(timestep)
    for event in super().events():
      self._events_subject.on_next(event)
    return timestep

  def reward_spec(self) -> Sequence[dm_env.specs.Array]:
    """See base class."""
    return self._env.reward_spec()

  def observation_spec(self) -> Sequence[Mapping[str, dm_env.specs.Array]]:
    """See base class."""
    return self._env.observation_spec()

  def action_spec(self) -> Sequence[dm_env.specs.DiscreteArray]:
    """See base class."""
    return self._env.action_spec()

  def close(self) -> None:
    """See base class."""
    super().close()
    self._action_subject.on_completed()
    self._timestep_subject.on_completed()
    self._events_subject.on_completed()

  def observables(self) -> SubstrateObservables:
    """Returns observables for the substrate."""
    return self._observables


def build_substrate(
    *,
    lab2d_settings: builder.Settings,
    individual_observations: Collection[str],
    global_observations: Collection[str],
    action_table: Sequence[Mapping[str, int]],
) -> Substrate:
  """Builds a Melting Pot substrate.

  Args:
    lab2d_settings: the lab2d settings for building the lab2d environment.
    individual_observations: names of the player-specific observations to make
      available to each player.
    global_observations: names of the dmlab2d observations to make available to
      all players.
    action_table: the possible actions. action_table[i] defines the dmlab2d
      action that will be forwarded to the wrapped dmlab2d environment for the
      discrete Melting Pot action i.

  Returns:
    The constructed substrate.
  """
  env = builder.builder(lab2d_settings)
  env = observables_wrapper.ObservablesWrapper(env)
  env = multiplayer_wrapper.Wrapper(
      env,
      individual_observation_names=individual_observations,
      global_observation_names=global_observations)
  env = discrete_action_wrapper.Wrapper(env, action_table=action_table)
  # Add a wrapper that augments adds an observation of the collective
  # reward (sum of all players' rewards).
  env = collective_reward_wrapper.CollectiveRewardWrapper(env)
  return Substrate(env)



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests of bots."""

from absl.testing import absltest
from absl.testing import parameterized

from meltingpot import bot
from meltingpot.testing import bots as test_utils


@parameterized.named_parameters((name, name) for name in bot.BOTS)
class BotTest(test_utils.BotTestCase):

  def test_step_without_error(self, name):
    factory = bot.get_factory(name)
    with factory.build() as policy:
      self.assert_compatible(
          policy,
          timestep_spec=factory.timestep_spec(),
          action_spec=factory.action_spec())


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for base wrapper."""

import inspect
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import dmlab2d

from meltingpot.utils.substrates.wrappers import base

_WRAPPED_METHODS = tuple([
    name for name, _ in inspect.getmembers(dmlab2d.Environment)
    if not name.startswith('_')
])


class WrapperTest(parameterized.TestCase):

  def test_instance(self):
    env = mock.Mock(spec_set=dmlab2d.Environment)
    wrapped = base.Lab2dWrapper(env=env)
    self.assertIsInstance(wrapped, dmlab2d.Environment)

  @parameterized.named_parameters(
      (name, name) for name in _WRAPPED_METHODS
  )
  def test_wrapped(self, method):
    env = mock.Mock(spec_set=dmlab2d.Environment)
    env_method = getattr(env, method)
    env_method.return_value = mock.sentinel

    wrapped = base.Lab2dWrapper(env=env)
    args = [object()]
    kwargs = {'a': object()}
    actual = getattr(wrapped, method)(*args, **kwargs)

    with self.subTest('args'):
      env_method.assert_called_once_with(*args, **kwargs)
    with self.subTest('return_value'):
      self.assertEqual(actual, env_method.return_value)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Collaborative Cooking.

A pure common interest cooking game inspired by Carroll et al. (2019) and
Strouse et al. (2021).

Carroll, M., Shah, R., Ho, M.K., Griffiths, T.L., Seshia, S.A., Abbeel, P. and
Dragan, A., 2019. On the utility of learning about humans for human-AI
coordination. arXiv preprint arXiv:1910.05789.

Strouse, D.J., McKee, K.R., Botvinick, M., Hughes, E. and Everett, R., 2021.
Collaborating with Humans without Human Data. arXiv preprint arXiv:2110.08176.
"""

import copy
from typing import Any, Dict, Mapping, Sequence

from ml_collections import config_dict as configdict

from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import game_object_utils
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

COOKING_TIME = 20
items = ["empty", "tomato", "dish", "soup"]

# Map a character to the prefab it represents in the ASCII map.
CHAR_PREFAB_MAP = {
    "P": "spawn_point",
    "#": "counter",
    "O": "tomato_dispenser",
    "D": "dish_dispenser",
    "T": "delivery_location",
    "C": "cooking_pot",
}

###########
# SPRITES #
###########

TARGET_SPRITE_SELF = {
    "name": "Self",
    "shape": shapes.CUTE_AVATAR,
    "palette": shapes.get_palette((50, 100, 200)),
    "noRotate": True,
}

BACKGROUND_LIGHT = (255, 255, 255, 255)
BACKGROUND_DARK = (82, 82, 82, 255)
OUTLINE = (85, 58, 23, 255)
OUTLINE_DARK = (49, 49, 49, 255)
INVISIBLE = (0, 0, 0, 0)
COUNTER = (115, 81, 39, 255)

SPRITES = {}
PALETTES = {}

SPRITES["interact"] = """
PPPPPPPPPPPPPPPP
PPPPPPPPPPPPPPPP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PP~~~~~~~~~~~~PP
PPPPPPPPPPPPPPPP
PPPPPPPPPPPPPPPP
"""

SPRITES["empty"] = """
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
"""
PALETTES["empty"] = {"~": INVISIBLE, "@": BACKGROUND_LIGHT}

SPRITES["counter"] = """
&&&&&&&&&&&&&&&&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&**************&
&&&&&&&&&&&&&&&&
"""
PALETTES["counter"] = {"*": COUNTER, "&": OUTLINE}

SPRITES["delivery_location"] = SPRITES["counter"]
PALETTES["delivery_location"] = {"*": BACKGROUND_DARK, "&": OUTLINE_DARK}

SPRITES["dish"] = """
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~++++~~~~~~
~~~~~+^^^^+~~~~~
~~~~~+^^^^+~~~~~
~~~~~+^^^^+~~~~~
~~~~~&++++&~~~~~
~~~~~~&&&&~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
"""
PALETTES["dish"] = {
    "~": [0, 0, 0, 0],
    "+": [255, 255, 255, 255],
    "^": [233, 239, 248, 255],
    "&": [221, 222, 238, 255]
}

SPRITES["soup"] = SPRITES["dish"]
PALETTES["soup"] = {
    "~": [0, 0, 0, 0],
    "+": [255, 255, 255, 255],
    "^": [236, 58, 74, 255],
    "&": [221, 222, 238, 255]
}

SPRITES["dish_dispenser"] = """
&&&&&&&&&&&&&&&&
&~~~~~~~~~~~~~~&
&~~~~~~~~~~~~~~&
&~~~~~~~~~~~~~~&
&~~~~~~~~~~~~~~&
&~~~~~++++~~~~~&
&~~~~+^^^^+~~~~&
&~~~~+^^^^+~~~~&
&~~~~+^^^^+~~~~&
&~~~~X++++X~~~~&
&~~~~~XXXX~~~~~&
&~~~~~~~~~~~~~~&
&~~~~~~~~~~~~~~&
&~~~~~~~~~~~~~~&
&~~~~~~~~~~~~~~&
&&&&&&&&&&&&&&&&
"""

PALETTES["dish_dispenser"] = {
    "&": OUTLINE_DARK,
    "~": BACKGROUND_DARK,
    "+": [255, 255, 255, 255],
    "^": [233, 239, 248, 255],
    "X": [221, 222, 238, 255]
}

SPRITES["tomato"] = """
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~++^+~~~~~~
~~~~~&O^---~~~~~
~~~~~O-----~~~~~
~~~~~O&-@--~~~~~
~~~~~~OO&&~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
"""
PALETTES["tomato"] = {
    "~": [0, 0, 0, 0],
    "+": [239, 81, 90, 255],
    "^": [29, 139, 43, 255],
    "&": [190, 53, 62, 255],
    "O": [151, 47, 52, 255],
    "-": [236, 58, 74, 255],
    "@": [240, 57, 75, 255]
}

SPRITES["tomato_dispenser"] = """
&&&&&&&&&&&&&&&&
&,,,,,,,,,,,,,,&
&,,,,,,,,,,,,,,&
&,,,,,,,,,,,,,,&
&,,,,,,,,,,,,,,&
&,,,,,,,,,,,,,,&
&,,,,,++^+,,,,,&
&,,,,XO^---,,,,&
&,,,,O-----,,,,&
&,,,,O&-@--,,,,&
&,,,,,OOXX,,,,,&
&,,,,,,,,,,,,,,&
&,,,,,,,,,,,,,,&
&,,,,,,,,,,,,,,&
&,,,,,,,,,,,,,,&
&&&&&&&&&&&&&&&&
"""

PALETTES["tomato_dispenser"] = {
    ",": BACKGROUND_DARK,
    "&": OUTLINE_DARK,
    "~": [0, 0, 0, 0],
    "+": [239, 81, 90, 255],
    "^": [29, 139, 43, 255],
    "X": [190, 53, 62, 255],
    "O": [151, 47, 52, 255],
    "-": [236, 58, 74, 255],
    "@": [240, 57, 75, 255]
}

SPRITES["cooking_pot_empty"] = """
&&&&&&&&&&&&&&&&
&~~~++++++++~~~&
&~~+^^^^^^^XO~~&
&~~+^^^^^^XXO~~&
&^^+^^^^^XXXO--&
&^~+^^^^XXXXO~-&
&^~+@@@@AAAAO~-&
&^^+@@@@AAAAO--&
&~~+@@@@AAAAO~~&
&~~@OOOOOOOO-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~~MMMMMMMM~~~&
&&&&&&&&&&&&&&&&
"""

SPRITES["cooking_pot_1"] = """
&&&&&&&&&&&&&&&&
&~~~++++++++~~~&
&~~+^^^^^^^XO~~&
&~~+^^^^^^XXO~~&
&^^+^^^^^XXXO--&
&^~+KKKKLLLLO~-&
&^~+KKKKLLLLO~-&
&^^+KKKKLLLLO--&
&~~+KKKKLLLLO~~&
&~~@OOOOOOOO-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~~MMMMMMMM~~~&
&&&&&&&&&&&&&&&&
"""

SPRITES["cooking_pot_2"] = """
&&&&&&&&&&&&&&&&
&~~~++++++++~~~&
&~~+^^^^^^^XO~~&
&~~+^^^^^^XXO~~&
&^^+KKKKKKLLO--&
&^~+KKKKLKLLO~-&
&^~+KLKKKKLLO~-&
&^^+KKKKKKLLO--&
&~~+KKKKKKLLO~~&
&~~@OOOOOOOO-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~~MMMMMMMM~~~&
&&&&&&&&&&&&&&&&
"""

SPRITES["cooking_pot_3"] = """
&&&&&&&&&&&&&&&&
&~~~++++++++~~~&
&~~+KKKKKKKKO~~&
&~~+KNKKKKKKO~~&
&^^+KKKKKKKKO--&
&^~+KKKKKKKKO~-&
&^~+KKKKKKKKO~-&
&^^+KKNKKKNKO--&
&~~+KKKKKKKKO~~&
&~~@OOOOOOOO-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~@MMMMMMMM-~~&
&~~~MMMMMMMM~~~&
&&&&&&&&&&&&&&&&
"""

PALETTES["cooking_pot"] = {
    "&": OUTLINE,
    "~": COUNTER,
    "+": [224, 231, 240, 255],
    "^": [140, 155, 181, 255],
    "X": [98, 95, 128, 255],
    "O": [238, 241, 241, 255],
    "-": [194, 206, 222, 255],
    "@": [92, 106, 135, 255],
    "A": [65, 66, 97, 255],
    "M": [139, 155, 181, 255],
    "K": [236, 58, 74, 255],
    "L": [161, 43, 43, 255],
    "N": [242, 226, 187, 255]
}

SPRITES["loading_bar"] = """
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
~~~abcdefghij~~~
~~~abcdefghij~~~
~~~~~~~~~~~~~~~~
~~~~~~~~~~~~~~~~
"""


def create_loading_bar_palette(count, finished=False):
  """Creates an incrementally colored loading bar based on count."""
  characters = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
  bars_palette = {"~": INVISIBLE}
  for idx in range(0, 10):
    if idx < count:
      if finished:
        bars_palette[characters[idx]] = (15, 188, 15, 255)
      else:
        bars_palette[characters[idx]] = (201, 178, 50, 255)
    else:
      bars_palette[characters[idx]] = (255, 255, 255, 255)

  return bars_palette


OFFSET_SIZE = 3
for s in ["empty", "tomato", "dish", "soup"]:
  SPRITES[s + "_offset"] = ("~~~~~~~~~~~~~~~~\n" * OFFSET_SIZE +
                            SPRITES[s][1:-OFFSET_SIZE*17])

###########
# PREFABS #
###########

SPAWN_POINT = {
    "name": "spawnPoint",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "layer": "logic",
                    "groups": ["spawnPoints"]
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}

# pylint: disable=g-complex-comprehension
INVENTORY = {
    "name": "inventory",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "wait",
                "stateConfigs": [{
                    "state": "wait"
                }] + [{
                    "state": item,
                    "sprite": item,
                    "layer": "overlay"
                } for item in items] + [{
                    "state": item + "_offset",
                    "sprite": item + "_offset",
                    "layer": "overlay"
                } for item in items]
            }
        },
        {
            "component": "Transform",
            "kwargs": {}
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": items + [item + "_offset" for item in items],
                "spriteShapes": ([SPRITES[item] for item in items] +
                                 [SPRITES[item + "_offset"] for item in items]),
                "palettes": ([PALETTES[item] for item in items] +
                             [PALETTES[item] for item in items]),
                "noRotates": [False]
            }
        },
        {
            "component": "Inventory",
            "kwargs": {
                "playerIndex": -1,  # Can be overwritten.
                "emptyState": "empty",
                "waitState": "wait"
            }
        },
    ]
}
# pylint: enable=g-complex-comprehension

loading_palettes = ([create_loading_bar_palette(idx) for idx in range(0, 10)] +
                    [create_loading_bar_palette(10, finished=True)])
LOADING_BAR = {
    "name": "loading_bar",
    "components": [{
        "component": "StateManager",
        "kwargs": {
            "initialState":
                "loading_bar_0",
            "stateConfigs": [{
                "state": "loading_bar_%d" % d, "layer": "overlay",
                "sprite": "loading_bar_%d" % d} for d in range(0, 11)],
        }
    }, {
        "component": "Transform",
        "kwargs": {}
    }, {
        "component": "Appearance",
        "kwargs": {
            "renderMode": "ascii_shape",
            "spriteNames": ["loading_bar_%d" % d for d in range(0, 11)],
            "spriteShapes": [SPRITES["loading_bar"] for _ in range(0, 11)],
            "palettes": loading_palettes,
            "noRotates": [True]
        }
    }, {
        "component": "LoadingBarVisualiser",
        "kwargs": {
            "totalTime": COOKING_TIME,
            "customStateNames": ["loading_bar_%d" % d for d in range(0, 11)],
        }
    }]
}


def create_base_prefab(name, layer="upperPhysical"):
  """Returns a base prefab with a given name on the given layer."""
  return {
      "name": f"{name}",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": f"{name}",
                  "stateConfigs": [{
                      "state": f"{name}",
                      "layer": layer,
                      "sprite": f"{name}",
                  }],
              }
          },
          {
              "component": "Transform",
              "kwargs": {}
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [name],
                  "spriteShapes": [SPRITES[name]],
                  "palettes": [PALETTES[name]],
                  "noRotates": [True]
              }
          }]
  }


def create_counter():
  """Returns a prefab which can contain one of any item."""
  base_prefab = create_base_prefab("counter")
  base_prefab["components"] += [{
      "component": "Container",
      "kwargs": {
          "reward": 0.0
      }
  }]
  return base_prefab


def create_dispenser(prefab_name, item_name):
  """Returns a prefab which dispenses items to avatars upon interaction."""
  base_prefab = create_base_prefab(prefab_name)
  base_prefab["components"] += [{
      "component": "Container",
      "kwargs": {
          "startingItem": item_name,
          "infinite": True,
          "reward": 0.0
      }
  }]
  return base_prefab


def create_receiver(prefab_name,
                    item_name,
                    reward=0,
                    global_reward=False):
  """Returns a prefab which can receive items from avatars.

  Args:
    prefab_name: the name of the prefab.
    item_name: the name of the accepted item.
    reward: value of reward given to avatar when object receives the item.
    global_reward: if true, reward all avatars.

  Returns:
    A prefab which can receive items.
  """
  base_prefab = create_base_prefab(prefab_name)
  base_prefab["components"] += [{
      "component": "Receiver",
      "kwargs": {
          "acceptedItems": item_name,
          "reward": reward,
          "globalReward": global_reward
      }
  }]
  return base_prefab


def create_cooking_pot(time_to_cook, reward=1):
  """Creates a cooking pot for tomatoes."""

  state_configs = []
  sprite_names = []
  pot_sprites = []
  cooking_pot_palettes = []
  custom_state_names = []
  available_foods = ["tomato"]

  # Create state for each combination of available foods.
  foods_in_pot = ["empty"] + available_foods
  for food1 in foods_in_pot:
    for food2 in foods_in_pot:
      for food3 in foods_in_pot:
        sprite_name = "CookingPot_%s_%s_%s" % (food1, food2, food3)
        name = "cooking_pot_%s_%s_%s" % (food1, food2, food3)
        entry = {"state": name,
                 "layer": "upperPhysical",
                 "sprite": sprite_name,
                 "groups": ["cooking_pot"]}
        state_configs.append(entry)
        sprite_names.append(sprite_name)
        pots_palette = PALETTES["cooking_pot"]
        cooking_pot_palettes.append(pots_palette)
        custom_state_names.append(name)

        if food1 == "empty":
          pot_sprites.append(SPRITES["cooking_pot_empty"])
        elif food2 == "empty":
          pot_sprites.append(SPRITES["cooking_pot_1"])
        elif food3 == "empty":
          pot_sprites.append(SPRITES["cooking_pot_2"])
        else:
          pot_sprites.append(SPRITES["cooking_pot_3"])

  # Create cooked state.
  sprite_name = "CookingPot_cooked"
  name = "cooking_pot_cooked"
  entry = {"state": "cooking_pot_cooked",
           "layer": "upperPhysical",
           "sprite": sprite_name,
           "groups": ["cooking_pot"]}
  state_configs.append(entry)
  sprite_names.append(sprite_name)
  pot_sprites.append(SPRITES["cooking_pot_3"])
  pots_palette = PALETTES["cooking_pot"]
  cooking_pot_palettes.append(pots_palette)
  custom_state_names.append(name)

  cooking_pot = {
      "name": "cooking_pot",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": (
                      "cooking_pot_empty_empty_empty"
                  ),
                  "stateConfigs": state_configs
              }
          },
          {
              "component": "Transform",
              "kwargs": {}
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": sprite_names,
                  "spriteShapes": pot_sprites,
                  "palettes": cooking_pot_palettes,
                  "noRotates": [True for _ in sprite_names]
              }
          },
          {
              "component": "CookingPot",
              "kwargs": {
                  "acceptedItems": available_foods,
                  "cookingTime": time_to_cook,
                  "reward": reward,
                  "customStateNames": custom_state_names
              }
          },
      ]
  }

  return cooking_pot


def create_prefabs(cooking_pot_pseudoreward: float = 0.0):
  """Creates a dictionary mapping names to template game objects."""
  prefabs = {
      "spawn_point": SPAWN_POINT,
      "inventory": INVENTORY,
      "loading_bar": LOADING_BAR,
      "counter": create_counter(),
      "dish_dispenser": create_dispenser(prefab_name="dish_dispenser",
                                         item_name="dish"),
      "tomato_dispenser": create_dispenser(prefab_name="tomato_dispenser",
                                           item_name="tomato"),
      "delivery_location": create_receiver(prefab_name="delivery_location",
                                           item_name="soup",
                                           reward=20,
                                           global_reward=True),
      "cooking_pot": create_cooking_pot(time_to_cook=COOKING_TIME,
                                        reward=cooking_pot_pseudoreward),
  }
  return prefabs

###########
# ACTIONS #
###########

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "interact": 0}
FORWARD    = {"move": 1, "turn":  0, "interact": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "interact": 0}
BACKWARD   = {"move": 3, "turn":  0, "interact": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "interact": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "interact": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "interact": 0}
INTERACT   = {"move": 0, "turn":  0, "interact": 1}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    INTERACT,
)

###########
# OBJECTS #
###########


def create_game_objects(ascii_map_string, prefabs):
  """Returns list of game objects from 'ascii_map' and 'char_prefab' mapping."""

  # Create all game objects.
  game_objects = []
  for char, _ in CHAR_PREFAB_MAP.items():
    transforms = game_object_utils.get_game_object_positions_from_map(
        ascii_map_string, char)
    for transform in transforms:
      # Add inventory game object for holding and visualising items.
      if char == "#" or char == "O" or char == "D":
        inventory_object = copy.deepcopy(prefabs["inventory"])
        go_transform = game_object_utils.get_first_named_component(
            inventory_object, "Transform")
        go_transform["kwargs"]["position"] = (transform.position.x,
                                              transform.position.y)
        game_objects.append(inventory_object)

      # Add loading bar object to cooking pots.
      if char == "C":
        loading_object = copy.deepcopy(prefabs["loading_bar"])
        go_transform = game_object_utils.get_first_named_component(
            loading_object, "Transform")
        go_transform["kwargs"]["position"] = (transform.position.x,
                                              transform.position.y)
        game_objects.append(loading_object)

  return game_objects


def create_avatar_object(player_idx: int,
                         target_sprite_self: Dict[str, Any]) -> Dict[str, Any]:
  """Create an avatar object that always sees itself as blue."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  # Setup the self vs other sprite mapping.
  source_sprite_self = "Avatar" + str(lua_index)
  custom_sprite_map = {source_sprite_self: target_sprite_self["name"]}

  interact_palette = {
      "P": colors.palette[player_idx] + (255,),
      "~": INVISIBLE,
  }

  live_state_name = "player{}".format(lua_index)
  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": live_state_name,
                  "stateConfigs": [
                      {"state": live_state_name,
                       "layer": "upperPhysical",
                       "sprite": source_sprite_self,
                       "contact": "avatar",
                       "groups": ["players"]},

                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [source_sprite_self],
                  "spriteShapes": [shapes.CUTE_AVATAR],
                  "palettes": [shapes.get_palette(colors.palette[player_idx])],
                  "noRotates": [True]
              }
          },
          {
              "component": "AdditionalSprites",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "customSpriteNames": [target_sprite_self["name"]],
                  "customSpriteShapes": [target_sprite_self["shape"]],
                  "customPalettes": [target_sprite_self["palette"]],
                  "customNoRotates": [target_sprite_self["noRotate"]],
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "spawnGroup": "spawnPoints",
                  "aliveState": live_state_name,
                  "waitState": "playerWait",
                  "speed": 1.0,
                  "actionOrder": ["move", "turn", "interact"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": 4},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "interact": {"default": 0, "min": 0, "max": 1},
                  },
                  "view": {
                      "left": 2,
                      "right": 2,
                      "forward": 3,
                      "backward": 1,
                      "centered": False
                  },
                  "spriteMap": custom_sprite_map,
              }
          },
          {
              "component": "InteractBeam",
              "kwargs": {
                  "cooldownTime": 1,
                  "shapes": [SPRITES["interact"]],
                  "palettes": [interact_palette]
              }
          },
          {"component": "AvatarCumulants",},
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })
    avatar_object["components"].append({
        "component": "AvatarMetricReporter",
        "kwargs": {
            "metrics": [
                {
                    "name": "ADDED_INGREDIENT_TO_COOKING_POT",
                    "type": "Doubles",
                    "shape": [],
                    "component": "AvatarCumulants",
                    "variable": "addedIngredientToCookingPot",
                },
                {
                    "name": "COLLECTED_SOUP_FROM_COOKING_POT",
                    "type": "Doubles",
                    "shape": [],
                    "component": "AvatarCumulants",
                    "variable": "collectedSoupFromCookingPot",
                },
            ]
        }
    })

  return avatar_object


def create_avatar_objects(num_players, prefabs):
  """Returns list of avatar objects of length 'num_players'."""
  game_objects = []
  for player_idx in range(0, num_players):
    lua_index = player_idx + 1
    game_object = create_avatar_object(player_idx,
                                       TARGET_SPRITE_SELF)
    game_objects.append(game_object)

    # Add inventory game object which will be connected to player at init.
    inventory_object = copy.deepcopy(prefabs["inventory"])
    game_object_utils.get_first_named_component(
        inventory_object,
        "Inventory")["kwargs"]["playerIndex"] = lua_index
    game_objects.append(inventory_object)

  return game_objects


def get_config():
  """Default configuration for training on the collaborative cooking level."""
  config = configdict.ConfigDict()

  # Cooking pot pseudoreward should be 0.0 for the canonical version of this
  # environment, but in order to train background bots it is sometimes useful
  # to give them a pseudoreward when they put items in the cooking pot. It has
  # the effect of shaping their behavior a bit in the right direction.
  config.cooking_pot_pseudoreward = 0.0

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  config.action_spec = specs.action(len(ACTION_SET))

  return config


def build(
    roles: Sequence[str],
    config: configdict.ConfigDict,
) -> Mapping[str, Any]:
  """Build the substrate given player roles."""
  num_players = len(roles)
  ascii_map = config.layout.ascii_map
  prefabs = create_prefabs(
      cooking_pot_pseudoreward=config.cooking_pot_pseudoreward)
  game_objects = create_game_objects(ascii_map, prefabs)
  extra_game_objects = create_avatar_objects(num_players, prefabs)
  game_objects += extra_game_objects
  # Build the rest of the substrate definition.
  substrate_definition = dict(
      levelName="collaborative_cooking",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      maxEpisodeLengthFrames=1000,
      spriteSize=8,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": ascii_map,
          "gameObjects": game_objects,
          "prefabs": prefabs,
          "charPrefabMap": CHAR_PREFAB_MAP,
      },
  )
  return substrate_definition



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for predator_prey__orchard.

Example video: https://youtu.be/gtd-ziZYJRI

See predator_prey.py for a detailed description applicable to all predator_prey
substrates.

In this variant there are two areas of the map containing food: an apple-rich
region to the north of the safe tall grass and an acorn-rich region to the east.
There are two possible prey strategies focusing on either apples or acorns.
However, in this case it is clear that focusing on acorns is the better
strategy since they are relatively close to the safe tall grass. They can easily
be collected and brought back to safety for consumption.
"""

from ml_collections import config_dict

from meltingpot.configs.substrates import predator_prey as base_config
from meltingpot.utils.substrates import specs

build = base_config.build

ASCII_MAP = """
/;__________,;_______,/
;]aa'X'XX''a|!a''''aA[,
!a''aaaaaa'X[]''aa&''A|
!X'aaAaaaaa''''aaaaa''|
!'&'aaaaaa''Aa'aaaaaa'|
!a'''X''''X'''a''''''a|
!aa''aaa''''''''''''aa|
L~+''aaa''=~~+XXXaA=~~J
;_]'''a'X'[_,L~~~~~J;_,
!XX'''a'X'''[_______]'|
!'''''a''''''XX'''''''|
!'r^^^^^^l'''X'A'''A''|
!'zv#****#^l'''''A''''|
!'''<#***##j'''A'''A''|
!''r###**#>''''''A'''X|
!''zv##***#^l''A'''A''|
!''''zvvvvvvj''''A'&''|
L+'''''''''''''''''''=J
/L~~~~~~~~~~~~~~~~~~~J/
"""

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
CHAR_PREFAB_MAP = {
    "*": {"type": "all", "list": ["safe_grass", "spawn_point_prey"]},
    "&": {"type": "all", "list": ["tiled_floor", "spawn_point_prey"]},
    "X": {"type": "all", "list": ["tiled_floor", "spawn_point_predator"]},
    "a": {"type": "all", "list": ["tiled_floor", "apple"]},
    "A": {"type": "all", "list": ["tiled_floor", "floor_acorn"]},
    ";": "nw_wall_corner",
    ",": "ne_wall_corner",
    "J": "se_wall_corner",
    "L": "sw_wall_corner",
    "_": "wall_north",
    "|": "wall_east",
    "~": "wall_south",
    "!": "wall_west",
    "=": "nw_inner_wall_corner",
    "+": "ne_inner_wall_corner",
    "]": "se_inner_wall_corner",
    "[": "sw_inner_wall_corner",
    "'": "tiled_floor",
    "#": "safe_grass",
    "<": "safe_grass_w_edge",
    "^": "safe_grass_n_edge",
    ">": "safe_grass_e_edge",
    "v": "safe_grass_s_edge",
    "l": "safe_grass_ne_corner",
    "j": "safe_grass_se_corner",
    "z": "safe_grass_sw_corner",
    "r": "safe_grass_nw_corner",
    "/": "fill",
}


def get_config():
  """Default configuration."""
  config = base_config.get_config()

  # Override the map layout settings.
  config.layout = config_dict.ConfigDict()
  config.layout.ascii_map = ASCII_MAP
  config.layout.char_prefab_map = CHAR_PREFAB_MAP

  # The specs of the environment (from a single-agent perspective).
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "STAMINA": specs.float64(),
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(152, 184),
  })

  # The roles assigned to each player.
  config.default_player_roles = ("predator",) * 5 + ("prey",) * 8

  return config



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for clean_up puppeteers."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized

from meltingpot.utils.puppeteers import clean_up
from meltingpot.utils.puppeteers import testutils

_NUM_COOPERATORS_KEY = 'num_cooperators'
_COOPERATE = mock.sentinel.cooperate
_DEFECT = mock.sentinel.defect


def _goals(puppeteer, num_cooperators, state=None):
  observations = [{_NUM_COOPERATORS_KEY: n} for n in num_cooperators]
  goals, state = testutils.goals_from_observations(puppeteer, observations,
                                                   state)
  return goals, state


class ConditionalCleanerTest(parameterized.TestCase):

  @parameterized.parameters(0, 1, 2)
  def test_niceness_period(self, niceness_period):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=1,
        threshold=100,
        reciprocation_period=1,
        niceness_period=niceness_period,
    )
    num_cooperators = [0, 0, 0]
    expected = [_COOPERATE] * niceness_period
    expected += [_DEFECT] * (len(num_cooperators) - niceness_period)
    actual, _ = _goals(puppeteer, num_cooperators)
    self.assertEqual(actual, expected)

  def test_reciprocation_trigger(self):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=1,
        threshold=4,
        reciprocation_period=1,
        niceness_period=0,
    )
    num_cooperators = [0, 1, 2, 3, 4]
    expected = [_DEFECT, _DEFECT, _DEFECT, _DEFECT, _COOPERATE]
    actual, _ = _goals(puppeteer, num_cooperators)
    self.assertSequenceEqual(actual, expected)

  @parameterized.parameters(1, 2)
  def test_reciprocation_period(self, reciprocation_period):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=1,
        threshold=1,
        reciprocation_period=reciprocation_period,
        niceness_period=0,
    )
    num_cooperators = [1, 0, 0, 0, 0]
    expected = [_COOPERATE] * reciprocation_period
    expected += [_DEFECT] * (len(num_cooperators) - reciprocation_period)
    actual, _ = _goals(puppeteer, num_cooperators)
    self.assertSequenceEqual(actual, expected)

  @parameterized.parameters(
      [(1, 0, 0, 1), (_DEFECT, _DEFECT, _DEFECT, _DEFECT)],
      [(1, 0, 1), (_DEFECT, _DEFECT, _COOPERATE)],
      [(1, 1), (_DEFECT, _COOPERATE)],
  )
  def test_recency_window(self, num_cooperators, expected):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=3,
        threshold=2,
        reciprocation_period=1,
        niceness_period=0,
    )
    actual, _ = _goals(puppeteer, num_cooperators)
    self.assertSequenceEqual(actual, expected)

  def test_niceness_persists(self):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=1,
        threshold=1,
        reciprocation_period=1,
        niceness_period=4,
    )
    num_cooperators = [1, 0, 0, 0, 0]
    expected = [_COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE, _DEFECT]
    actual, _ = _goals(puppeteer, num_cooperators)
    self.assertSequenceEqual(actual, expected)

  def test_reciprocation_extends_niceness(self):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=1,
        threshold=1,
        reciprocation_period=4,
        niceness_period=2,
    )
    num_cooperators = [1, 0, 0, 0, 0]
    expected = [_COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE, _DEFECT]
    actual, _ = _goals(puppeteer, num_cooperators)
    self.assertSequenceEqual(actual, expected)

  def test_reciprocation_extends_reciprocation(self):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=1,
        threshold=1,
        reciprocation_period=3,
        niceness_period=0,
    )
    num_cooperators = [1, 1, 0, 0, 0]
    expected = [_COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE, _DEFECT]
    actual, _ = _goals(puppeteer, num_cooperators)
    self.assertSequenceEqual(actual, expected)

  def test_resets_on_first(self):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=8,
        threshold=1,
        reciprocation_period=8,
        niceness_period=1,
    )
    _, state = _goals(puppeteer, [0, 0, 1, 0])
    num_cooperators = [0, 0, 0, 0]
    expected = [_COOPERATE, _DEFECT, _DEFECT, _DEFECT]
    actual, _ = _goals(puppeteer, num_cooperators, state)
    self.assertSequenceEqual(actual, expected)

  def test_impulse_response(self):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=4,
        threshold=1,
        reciprocation_period=2,
        niceness_period=0,
    )
    num_defections = [1, 0, 0, 0, 0, 0]
    expected = [
        _COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE, _DEFECT
    ]
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

  def test_boxcar_response(self):
    puppeteer = clean_up.ConditionalCleaner(
        clean_goal=_COOPERATE,
        eat_goal=_DEFECT,
        coplayer_cleaning_signal=_NUM_COOPERATORS_KEY,
        recency_window=4,
        threshold=1,
        reciprocation_period=2,
        niceness_period=0,
    )
    num_defections = [1, 1, 1, 0, 0, 0, 0, 0]
    expected = [
        _COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE, _COOPERATE,
        _COOPERATE, _DEFECT
    ]
    actual, _ = _goals(puppeteer, num_defections)
    self.assertSequenceEqual(actual, expected)

if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for predator_prey__alley_hunt.

Example video: https://youtu.be/ctVjhn7VYgo

See predator_prey.py for a detailed description applicable to all predator_prey
substrates.

In this variant prey must forage for apples in a maze with many dangerous
dead-end corridors where they could easily be trapped by predators.
"""

from ml_collections import config_dict

from meltingpot.configs.substrates import predator_prey as base_config
from meltingpot.utils.substrates import specs

build = base_config.build

ASCII_MAP = """
;________________________,
!aa''''''''''''''''''''aa|
!a'''''''''a''=+''''''''a|
!''=~~~+''=+''|!''=~~~+''|
!''[__,!''|!''|!''[___]''|
!''''a|!''|!aa|!'''''''''|
!''=~~J!''|L~~J!'a'=~~~+'|
!''|///!''[____]'a'|///!a|
!''|///!'''''''''''[__,L~J
!''[___]'XX''''X''''<*[__,
!''''''''''a''''XX''<****|
!'aa'''X''''''a'''XX<****|
!''''''''''a''''XX''<****|
!''=~~~+'''''''X''''<*=~~J
!''|///!'XX''''''''=~~J;_,
!''|///!''=~~~~+'a'|///!a|
!''[__,!''|;__,!'a'[___]'|
!''''a|!''|!aa|!'''''''''|
!''=~~J!''|!''|!''=~~~+''|
!''[___]''[]''|!''[___]''|
!a'''''''''a''[]''''''''a|
!aa''''''''''''''''''''aa|
L~~~~~~~~~~~~~~~~~~~~~~~~J
"""

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
CHAR_PREFAB_MAP = {
    "*": {"type": "all", "list": ["safe_grass", "spawn_point_prey"]},
    "X": {"type": "all", "list": ["tiled_floor", "spawn_point_predator"]},
    "a": {"type": "all", "list": ["tiled_floor", "apple"]},
    ";": "nw_wall_corner",
    ",": "ne_wall_corner",
    "J": "se_wall_corner",
    "L": "sw_wall_corner",
    "_": "wall_north",
    "|": "wall_east",
    "~": "wall_south",
    "!": "wall_west",
    "=": "nw_inner_wall_corner",
    "+": "ne_inner_wall_corner",
    "]": "se_inner_wall_corner",
    "[": "sw_inner_wall_corner",
    "'": "tiled_floor",
    "<": "safe_grass_w_edge",
    ">": "safe_grass",
    "/": "fill",
}


def get_config():
  """Default configuration."""
  config = base_config.get_config()

  # Override the map layout settings.
  config.layout = config_dict.ConfigDict()
  config.layout.ascii_map = ASCII_MAP
  config.layout.char_prefab_map = CHAR_PREFAB_MAP

  # The specs of the environment (from a single-agent perspective).
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "STAMINA": specs.float64(),
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(184, 208),
  })

  # The roles assigned to each player.
  config.default_player_roles = ("predator",) * 5 + ("prey",) * 8

  return config



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Puppeteers for running_with_scissors_in_the_matrix."""

from meltingpot.utils.puppeteers import in_the_matrix


class CounterPrevious(in_the_matrix.RespondToPrevious):
  """Puppeteer for a running with scissors bot.

  This bot will always play the best response strategy to whatever its
  partner played in the previous interaction. So if its partner last played
  rock then it will play paper. If its partner last played paper then it will
  play scissors. If its partner last played scissors then it will play rock.

  Important note: this puppeteer does not discriminate between coplayers. So it
  only makes sense in two-player substrates (e.g.
  `running_with_scissors_in_the_matrix__repeated`).
  """

  def __init__(
      self,
      rock_resource: in_the_matrix.Resource,
      paper_resource: in_the_matrix.Resource,
      scissors_resource: in_the_matrix.Resource,
      margin: int,
  ) -> None:
    """Initializes the puppeteer.

    Args:
      rock_resource: The rock resource.
      paper_resource: The paper resource.
      scissors_resource: The scissors resource.
      margin: Try to collect `margin` more of the target resource than the other
        resource before interacting.
    """
    responses = {
        rock_resource: paper_resource,
        paper_resource: scissors_resource,
        scissors_resource: rock_resource,
    }
    super().__init__(responses, margin)
